from .agent import JAImsAgent
from .openai_wrappers import (
    JAImsGPTModel,
    JAImsTokensExpense,
    estimate_token_count,
    JAImsOpenaiKWArgs,
    JAImsOptions,
)

# TODO: remove stars and export only what's necessary
from .exceptions import *
from .function_handler import *
from .transaction_storage import JAImsTransactionStorageInterface



class JAImsTransactionStorageInterface(object):
    """
    Interface for storing JAIms transactions.
    Override this class to implement your own storage, will store a pair of openai request and response.
    """

    def store_transaction(self, request: dict, response: dict):
        pass



from __future__ import annotations
from enum import Enum
import json
from typing import Any, List, Dict, Optional, Callable

from jaims.exceptions import JAImsUnexpectedFunctionCall


# Enum class over all Json Types
class JAImsJsonSchemaType(Enum):
    STRING = "string"
    NUMBER = "number"
    OBJECT = "object"
    ARRAY = "array"
    BOOLEAN = "boolean"
    NULL = "null"


class JAImsParamDescriptor:
    """
    Describes a parameter to be used in the OPENAI API.

    Attributes
    ----------
        name : str
            the parameter name
        description : str
            the parameter description
        json_type : JsonType:
            the parameter json type
        attributes_params_descriptors : list of JAImsParamDescriptor
            the list of parameters descriptors for the attributes of the parameter
            in case the parameter is an object, defualts to None
        array_type_descriptors : list of JAImsParamDescriptor
            the parameter descriptors for the array type in case the parameter is an array, defaults to None
        enum_values:
            the list of values in case the parameter is an enum, defaults to None
        required : bool
            whether the parameter is required or not, defaults to True

    """

    def __init__(
        self,
        name: str,
        description: str,
        json_type: JAImsJsonSchemaType,
        attributes_params_descriptors: Optional[List[JAImsParamDescriptor]] = None,
        array_type_descriptors: Optional[List[JAImsParamDescriptor]] = None,
        array_type_any_valid: bool = True,
        enum_values: Optional[List[Any]] = None,
        required: bool = True,
    ):
        self.name = name
        self.description = description
        self.json_type = json_type
        self.attributes_params_descriptors = attributes_params_descriptors
        self.array_type_descriptors = array_type_descriptors
        self.array_type_any_valid = array_type_any_valid
        self.enum_values = enum_values
        self.required = required

    def get_jsonapi_schema(self) -> Dict[str, Any]:
        """
        Returns the jsonapi schema for the parameter.
        """
        schema: dict[str, Any] = {
            "type": self.json_type.value,
            "description": self.description,
        }

        if (
            self.json_type == JAImsJsonSchemaType.OBJECT
            and self.attributes_params_descriptors
        ):
            schema["properties"] = {}
            schema["required"] = []
            for param in self.attributes_params_descriptors:
                schema["properties"][param.name] = param.get_jsonapi_schema()
                if param.required:
                    schema["required"].append(param.name)

        if self.json_type == JAImsJsonSchemaType.ARRAY and self.array_type_descriptors:
            items_schema = [
                desc.get_jsonapi_schema() for desc in self.array_type_descriptors
            ]
            if self.array_type_any_valid:
                schema["items"] = {"anyOf": items_schema}
            else:
                schema["items"] = [items_schema]

        if self.enum_values:
            schema["enum"] = self.enum_values

        return schema


class JAImsFuncWrapper:
    """
    Wraps a function call to be used in the OPENAI API.
    Holds the function to be called as well as the function name, its description that the agent will understand,
    and the response json schema.

    Attributes
    ----------
        function : Callable[..., Any]
            the function to be called
        name : str
            the function name
        description : str
            the function description
        params_descriptors: List[JAImsParamDescriptor]
            the list of parameters descriptors

    Methods
    -------
        call(params: Dict[str, Any]) -> Any
            calls the function with the given parameters
        get_jsonapi_schema() -> Dict[str, Any]
            returns the jsonapi schema for the function
    """

    def __init__(
        self,
        function: Callable[..., Any],
        name: str,
        description: str,
        params_descriptors: List[JAImsParamDescriptor],
    ):
        self.function = function
        self.name = name
        self.description = description
        self.params_descriptors = params_descriptors

    def call(self, params: Dict[str, Any]) -> Any:
        """
        Calls the function with the given parameters.

        Parameters
        ----------
            params : dict
                the parameters to be passed to the function
        """
        return self.function(**params)

    def get_jsonapi_schema(self) -> Dict[str, Any]:
        """
        Returns the jsonapi schema for the function.
        """
        schema = {
            "type": "object",
            "properties": {},
            "required": [],
        }

        for param in self.params_descriptors:
            schema["properties"][param.name] = param.get_jsonapi_schema()
            if param.required:
                schema["required"].append(param.name)

        return schema


class JAImsFunctionHandler:
    """
    Handles the functions to be used in the OPENAI API.
    Holds a list of JAImsFuncWrapper instances.

    Attributes
    ----------
        functions : List[JAImsFuncWrapper]
            the list of functions to be called
    """

    def handle_from_message(
        self, message: Dict[str, Any], functions: List[JAImsFuncWrapper]
    ) -> List[Dict[str, Any]]:
        """
        Handles a function_call message, calling the appropriate function.

        Parameters
        ----------
            message : dict
                the message from the agent

        Returns
        -------
            dict
                the function result message to be sent to openai

        Raises
        ------
            UnexpectedFunctionCallException
                if the function name is not found in the functions list
        """
        tool_calls = message.get("tool_calls", [])
        function_calls = []
        for tool_call in tool_calls:
            tool_call_id = tool_call["id"]
            function_name = tool_call["function"]["name"]
            function_args = tool_call["function"]["arguments"]
            function_calls.append(
                {
                    "tool_call_id": tool_call_id,
                    "name": function_name,
                    "args": json.loads(function_args) if function_args else {},
                }
            )

        # invoke function
        return self.__call_functions(function_calls, functions)

    def __call_functions(
        self, function_calls: List[dict], functions
    ) -> List[dict[str, Any]]:
        # Check if function_name exists in functions, if not, raise UnexpectedFunctionCallException

        results = []
        for fc in function_calls:
            function_name = fc["name"]
            function_wrapper = next(
                (f for f in functions if f.name == function_name), None
            )
            if not function_wrapper:
                raise JAImsUnexpectedFunctionCall(function_name)

            try:
                fc_result = function_wrapper.function(**fc["args"])
                results.append(
                    {
                        "name": function_name,
                        "tool_call_id": fc["tool_call_id"],
                        "role": "tool",
                        "content": json.dumps(fc_result),
                    }
                )
            except TypeError as e:
                results.append(
                    {
                        "name": function_name,
                        "tool_call_id": fc["tool_call_id"],
                        "role": "tool",
                        "content": json.dumps({"error": str(e)}),
                    }
                )

        # If the name of the current function matches the provided name
        # Call the function and return its result
        return results


def parse_function_wrappers_to_tools(
    functions: List[JAImsFuncWrapper],
) -> List[Dict[str, Any]]:
    openai_functions = []
    for function in functions:
        function_data = {
            k: v
            for k, v in {
                "name": function.name,
                "description": function.description,
                "parameters": function.get_jsonapi_schema(),
            }.items()
            if v is not None
        }

        openai_functions.append({"type": "function", "function": function_data})

    return openai_functions



from io import BytesIO
from typing import List, Optional
from jaims.openai_wrappers import (
    estimate_token_count,
    JAImsOptions,
    JAImsOpenaiKWArgs,
)

from jaims.exceptions import JAImsTokensLimitExceeded
import json

from jaims.function_handler import parse_function_wrappers_to_tools
from math import ceil
import base64
from PIL import Image


class HistoryManager:
    """
    Manages the history of messages sent to OpenAI.
    The history manager stores the messages in memory and, when the agent is set to optimize the context,
    it optimizes the history to fit the max tokens supported by the current llm model.

    Parameters
    ----------
        history : list (optional)
            the history of messages to be sent to openai
    """

    def __init__(
        self,
        history: Optional[List] = None,
    ):
        """
        Returns a new HistoryManager instance.

        Parameters
        ----------
            history : list (optional)
                the initial history of messages to be sent to openai
        """
        self.__history = history or []

    def add_messages(self, messages: List):
        """
        Pushes new messages in the history.

        Parameters
        ----------
            message : str
                the message to be added
        """

        if not all(isinstance(message, dict) for message in messages):
            raise TypeError(
                "All messages must be dicts, conforming to OpenAI API specification."
            )

        # this workaround is necessary because openai made a mess
        # with the new return types of the openai api, that were just a plain dictionary before and
        # now are a thousand classes often identical to each other.
        # I'm adding content because when I do the model_dump() of the model, None values are skipped
        # but "content" must be passed to none otherwise the api breaks.
        for message in messages:
            if "content" not in message:
                message["content"] = None

        self.__history.extend(messages)

    def get_messages_for_current_run(
        self,
        options: JAImsOptions,
        openai_kwargs: JAImsOpenaiKWArgs,
    ) -> List:
        """
        Returns the messages to be sent to openai for the current run.

        Parameters
        ----------
            options : JAImsOptions
                the options for the current run
            openai_kwargs : JAImsOpenaiKWArgs
                the openai kwargs for the current run

        Raises
        ------
            TokensLimitExceeded
                if the max tokens to be used exceed the max tokens supported by the current llm model

        Developer Notes
        ---------------

        # Optimization Feature
        This is roughly how the optimization feature works right now, plus some notes on how I intend to improve it.
        The tokens for the messages to send to openai are calculated by composing the mandatory_context, the chat history managed by this class
        and the functions in a variable named compound_history.
        They are parsed to a json string and tiktoken evaluates the token consumption based on the current llm model setting.

        Right now I'm parsing the whole compound_history, including the functions, to json and calculating the tokens based on the json string,
        but I'm not sure how accurate this is.
        On rough estimates made with the tokenizer from openai they seem to be pretty accurate, but I need to investigate more.

        A high level overview of the optimization loop is the following:

        the max_tokens to be used are the max tokens supported by the current llm model, minus the tokens to leave out for the response
        from openai passed in agent_max_tokens (passed with a default value) .

        The compound_history tokens are calculated, and while they exceed the context max_tokens:
        1. the first (oldest) message from the chat history between the user and the angent is popped
        2. the compound_history tokens are recalculated
        3. if the compound_history tokens are still above the context max_tokens, the process is repeated from step 1
            3.1. if the chat history between user and agent remains empty for some reason
                (this could happen for instance if functions and mandatory context are way too big),
                an exception is raised. I think it aids development but have to think about it.
        4. if the compound_history tokens are below the context max_tokens, the mandatory_context + the optimized history are returned.

        TODO:
        - optimize functions messages in the history

        MAYBE TODO:
        - it would be nice to have an auto-scale up of the context, for instance passing from the gpt-3.5-turbo 4k to the 16k model.
        """

        if not options or not openai_kwargs:
            raise ValueError("options and openai_kwargs must be provided.")

        # Copying the whole history to avoid altering the original one
        history_buffer = self.__history.copy()

        # If last_n_turns is set, only keep the last n messages
        if options.message_history_size is not None:
            history_buffer = history_buffer[-options.message_history_size :]

        # create the compound history with the mandatory context
        # the actual chat history and the functions to calculate the tokens
        json_functions = parse_function_wrappers_to_tools(openai_kwargs.tools or [])
        leading_prompts = options.leading_prompts or []
        trailing_prompts = options.trailing_prompts or []
        compound_history = (
            leading_prompts + history_buffer + (json_functions) + trailing_prompts
        )

        # the max tokens to be used are the max tokens supported by the current
        # openai model minus the tokens to leave out for the response from openai
        context_max_tokens = openai_kwargs.model.max_tokens - openai_kwargs.max_tokens

        # calculate the tokens for the compound history
        messages_tokens = self.__tokens_from_messages(
            compound_history, openai_kwargs.model
        )

        if options.optimize_context:
            while messages_tokens > context_max_tokens:
                if not history_buffer:
                    raise JAImsTokensLimitExceeded(
                        openai_kwargs.model.max_tokens,
                        messages_tokens,
                        openai_kwargs.max_tokens,
                        has_optimized=True,
                    )

                # Popping the first (oldest) message from the chat history between the user and agent
                history_buffer.pop(0)

                # Recalculating the tokens for the compound history
                messages_tokens = self.__tokens_from_messages(
                    leading_prompts
                    + history_buffer
                    + json_functions
                    + trailing_prompts,
                    openai_kwargs.model,
                )
        elif messages_tokens > context_max_tokens:
            raise JAImsTokensLimitExceeded(
                openai_kwargs.model.max_tokens,
                messages_tokens,
                openai_kwargs.max_tokens,
                has_optimized=False,
            )

        llm_messages = leading_prompts + history_buffer + trailing_prompts

        return llm_messages

    def clear_history(self):
        """
        Clears the history.
        """
        self.__history = []

    def get_history(self):
        """
        Returns entire history.
        """
        return self.__history

    def __tokens_from_messages(self, messages: List, model):
        """Returns the number of tokens used by a list of messages."""

        images = []
        parsed = []
        for message in messages:
            message_copy = message.copy()

            if isinstance(message.get("content", None), list):
                filtered_content = []
                for item in message["content"]:
                    if (
                        isinstance(item, dict)
                        and item.get("image_url", None)
                        and item["image_url"]["url"].startswith(
                            "data:image/jpeg;base64,"
                        )
                    ):
                        images.append(
                            item["image_url"]["url"].replace(
                                "data:image/jpeg;base64,", ""
                            )
                        )
                    else:
                        filtered_content.append(item)
                message_copy["content"] = filtered_content
            parsed.append(message_copy)

        image_tokens = 0
        for image in images:
            width, height = self.__get_image_size_from_base64(image)
            image_tokens += self.__count_image_tokens(width, height)

        return estimate_token_count(json.dumps(parsed), model) + image_tokens

    def __get_image_size_from_base64(self, base64_string):
        image_data = base64.b64decode(base64_string)
        image = Image.open(BytesIO(image_data))

        return image.size

    def __count_image_tokens(self, width: int, height: int):
        h = ceil(height / 512)
        w = ceil(width / 512)
        n = w * h
        total = 85 + 170 * n
        return total



import json
import logging
import os
from typing import Any, Generator, List, Optional, Union

from jaims.transaction_storage import (
    JAImsTransactionStorageInterface,
)

import openai
from openai.types.chat import ChatCompletionChunk, ChatCompletion

from jaims.openai_wrappers import (
    JAImsGPTModel,
    JAImsTokensExpense,
    JAImsOpenaiKWArgs,
    JAImsOptions,
    estimate_token_count,
    get_openai_response,
)
from jaims.exceptions import (
    JAImsMissingOpenaiAPIKeyException,
    JAImsMaxConsecutiveFunctionCallsExceeded,
)
from jaims.function_handler import (
    JAImsFunctionHandler,
)
from jaims.history_manager import HistoryManager


class JAImsCallContext:
    """
    Represents the context for a JAIms run.
    It is used when in case of function calling, the run restarts recursively.

    Args:
        openai_kwargs (JAImsOpenaiKWArgs): The OpenAI keyword arguments.
        options (JAImsOptions): The JAIms options.
        iterations (int, optional): The number of iterations. Defaults to 0.
    """

    def __init__(
        self,
        openai_kwargs: JAImsOpenaiKWArgs,
        options: JAImsOptions,
        iterations: int = 0,
    ):
        self.openai_kwargs = openai_kwargs
        self.options = options
        self.iterations = iterations

    def add_iteration(self):
        self.iterations += 1
        if self.iterations > self.options.max_consecutive_function_calls:
            raise JAImsMaxConsecutiveFunctionCallsExceeded(
                f"Max consecutive function calls exceeded ({self.options.max_consecutive_function_calls})"
            )


class JAImsAgent:
    """
    JAImsAgent realizes the class that interacts with the OpenAI API. It is an agent capable of
    tool calling, with built in history management and token expense tracking.

    Attributes:
        openai_kwargs (JAImsOpenaiKWArgs): The OpenAI keyword arguments.
        options (JAImsOptions): The options for the agent.
        openai_api_key (Optional[str]): The OpenAI API key.
        transaction_storage (JAImsTransactionStorageInterface): The transaction storage interface.

    Methods:
        run: Runs the agent with the given messages and options.
        get_expenses: Returns the tokens spent in the current session (all runs performed on this agent).
        get_last_run_expenses: Returns the tokens spent in the last run of the agent.
        get_run_history: Returns the history that will be sent to OpenAI for a run (returns the messages history passed to openai in the last run).
        get_history: Returns the complete history of the agent.
        clear_history: Clears the history.
    """

    def __init__(
        self,
        openai_kwargs: JAImsOpenaiKWArgs = JAImsOpenaiKWArgs(),
        options: JAImsOptions = JAImsOptions(),
        openai_api_key: Optional[str] = None,
        transaction_storage: Optional[JAImsTransactionStorageInterface] = None,
    ):
        openai_api_key = openai_api_key or os.getenv("OPENAI_API_KEY")
        if not openai_api_key:
            raise JAImsMissingOpenaiAPIKeyException()
        openai.api_key = openai_api_key

        self.__openai_kwargs = openai_kwargs
        self.__options = options
        self.__expense = JAImsAgent.__init_expense_dictionary()
        self.__last_run_expense = JAImsAgent.__init_expense_dictionary()
        self.__function_handler = JAImsFunctionHandler()
        self.__history_manager = HistoryManager()
        if transaction_storage is None:
            self.__transaction_storage = JAImsTransactionStorageInterface()
        else:
            self.__transaction_storage = transaction_storage

    def run(
        self,
        messages: Optional[List[dict]] = None,
        override_options: Optional[JAImsOptions] = None,
        override_openai_kwargs: Optional[JAImsOpenaiKWArgs] = None,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Starts a run calling openai with the passed messages.
        During a run, unless cleared, the history of the previous runs is preserved and optimized (see HistoryManager), unless clear_history is called.

        Args:
            messages (Optional[List[dict]], optional): The messages to be sent to OpenAI.
            override_options (Optional[JAImsOptions], optional): The options to be used for this run (entirely overriding those passed in constructor, only for this run).
            override_openai_kwargs (Optional[JAImsOpenaiKWArgs], optional): The OpenAI keyword arguments to be used for this run (entirely overriding those passed in constructor, only for this run).
        """
        messages = messages or []
        if override_openai_kwargs:
            messages = override_openai_kwargs.messages

        self.__history_manager.add_messages(messages)
        options = override_options or self.__options
        openai_kwargs = override_openai_kwargs or self.__openai_kwargs
        call_context = JAImsCallContext(openai_kwargs, options)

        self.__last_run_expense = JAImsAgent.__init_expense_dictionary()
        return self.__call_openai(call_context)

    def __call_openai(
        self, call_context: JAImsCallContext
    ) -> Union[str, Generator[str, None, None]]:
        # throws if max consecutive calls is exceeded
        call_context.add_iteration()
        messages = self.__history_manager.get_messages_for_current_run(
            options=call_context.options,
            openai_kwargs=call_context.openai_kwargs,
        )

        call_context.openai_kwargs.messages = messages

        response = get_openai_response(
            call_context.openai_kwargs,
            call_context.options,
        )

        if isinstance(response, openai.Stream):
            return self.__handle_streaming_response(response, call_context)
        else:
            return self.__handle_response(response, call_context)

    def __handle_streaming_response(
        self,
        response: openai.Stream[ChatCompletionChunk],
        call_context: JAImsCallContext,
    ) -> Generator[str, None, None]:
        accumulated_chunks = None
        for completion_chunk in response:
            if len(completion_chunk.choices) > 0:
                # check content exists
                message_delta = completion_chunk.choices[0].delta
                accumulated_chunks = JAImsAgent.__accumulate_choice_delta(
                    accumulated_chunks, message_delta
                )

                if (
                    call_context.options.debug_stream_function_call
                    and message_delta.tool_calls
                ):
                    print(
                        # type: ignore
                        message_delta.tool_calls[0].function.arguments,
                        flush=True,
                        end="",
                    )

                if completion_chunk.choices[0].finish_reason is not None:
                    if (
                        call_context.options.debug_stream_function_call
                        and accumulated_chunks.tool_calls
                    ):
                        print("\n")

                    # rebuilding entire completion chunk with accumulated delta
                    completion_chunk.choices[0].delta = accumulated_chunks
                    self.__store_transaction(call_context, completion_chunk)

                    self.__handle_token_expense_from_openai_response(
                        accumulated_chunks.model_dump(), call_context
                    )

                    yield from self.__handle_response_message(
                        accumulated_chunks, call_context
                    )

                if message_delta.content:
                    yield message_delta.content

    def __handle_response(
        self, response: ChatCompletion, call_context: JAImsCallContext
    ):
        if len(response.choices) == 0:
            return ""

        # log token expense
        self.__handle_token_expense_from_openai_response(
            response.model_dump(), call_context
        )

        self.__store_transaction(call_context, response)

        message = response.choices[0].message
        return self.__handle_response_message(message, call_context)

    def __handle_response_message(self, message, call_context: JAImsCallContext):
        logger = logging.getLogger(__name__)
        logger.debug(f"OpenAI response message:\n{message}")

        message_dict = message.model_dump(exclude_none=True)
        self.__history_manager.add_messages([message_dict])

        if message.tool_calls:
            result_messages = self.__function_handler.handle_from_message(
                message=message_dict,
                functions=call_context.openai_kwargs.tools or [],
            )
            if call_context.openai_kwargs.tool_choice == "auto":
                self.__history_manager.add_messages(result_messages)
                return self.__call_openai(call_context)

        # if the response was streaming only an empty string must be returned
        # because the streaming generator already yielded the response, otherwise
        # the content is returned.
        if call_context.openai_kwargs.stream:
            return ""
        else:
            return message.content or ""

    def __store_transaction(
        self,
        call_context: JAImsCallContext,
        response: Union[ChatCompletion, ChatCompletionChunk],
    ):
        self.__transaction_storage.store_transaction(
            request=call_context.openai_kwargs.to_dict(),
            response=response.model_dump(exclude_none=True),
        )

    def __handle_token_expense_from_openai_response(
        self, response, call_context: JAImsCallContext
    ):
        if not call_context.openai_kwargs.stream:
            expense = JAImsTokensExpense.from_openai_usage_dictionary(
                call_context.openai_kwargs.model, response["usage"]
            )
        else:
            sent_messages = self.__history_manager.get_messages_for_current_run(
                options=call_context.options,
                openai_kwargs=call_context.openai_kwargs,
            )
            prompt_tokens = estimate_token_count(
                json.dumps(sent_messages), model=call_context.openai_kwargs.model
            )
            completion_tokens = estimate_token_count(
                json.dumps(response), model=call_context.openai_kwargs.model
            )
            total_tokens = prompt_tokens + completion_tokens
            expense = JAImsTokensExpense(
                gpt_model=call_context.openai_kwargs.model,
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens,
                rough_estimate=True,
            )

        self.__last_run_expense[expense.gpt_model.string].add_from(expense)
        self.__expense[expense.gpt_model.string].add_from(expense)

    def get_expenses(self):
        """
        Returns the tokens spent in the current session in an array, one for each model.
        """
        return [
            expense for expense in self.__expense.values() if expense.total_tokens > 0
        ]

    def get_last_run_expenses(self):
        """
        Returns the tokens spent in the last run of the agent in an array, one for each model.
        """
        return [
            expense
            for expense in self.__last_run_expense.values()
            if expense.total_tokens > 0
        ]

    def get_run_history(
        self,
        override_options: Optional[JAImsOptions] = None,
        override_openai_kwargs: Optional[JAImsOpenaiKWArgs] = None,
    ):
        """
        Returns the history that will be sent given the current options and openai kwargs to openai for a run.
        """
        options = override_options or self.__options
        openai_kwargs = override_openai_kwargs or self.__openai_kwargs

        messages = self.__history_manager.get_messages_for_current_run(
            options, openai_kwargs
        )
        return messages

    def get_history(self):
        """
        Returns the complete history of the agent.
        """

        return self.__history_manager.get_history()

    def clear_history(self):
        """
        Clears the history.
        """
        self.__history_manager.clear_history()

    @staticmethod
    def __init_expense_dictionary():
        dict = {}
        for gpt in JAImsGPTModel:
            dict[gpt.string] = JAImsTokensExpense(gpt_model=gpt)

        return dict

    @staticmethod
    def __merge_tool_calls(existing_tool_calls, new_tool_calls_delta):
        if not existing_tool_calls:
            return new_tool_calls_delta

        new_tool_calls = existing_tool_calls[:]
        for new_call_delta in new_tool_calls_delta:
            existing_call = next(
                (item for item in new_tool_calls if item.index == new_call_delta.index),
                None,
            )
            if not existing_call:
                new_tool_calls.append(new_call_delta)
            else:
                if (
                    existing_call.type != new_call_delta.type
                    and new_call_delta.type is not None
                ):
                    existing_call.type = new_call_delta.type
                if (
                    existing_call.id != new_call_delta.id
                    and new_call_delta.id is not None
                ):
                    existing_call.id = new_call_delta.id
                if existing_call.function is None:
                    existing_call.function = new_call_delta.function
                else:
                    if (
                        existing_call.function.name != new_call_delta.function.name
                        and new_call_delta.function.name is not None
                    ):
                        existing_call.function.name = new_call_delta.function.name
                    existing_call.function.arguments = (
                        existing_call.function.arguments or ""
                    ) + (new_call_delta.function.arguments or "")

        return new_tool_calls

    @staticmethod
    def __accumulate_choice_delta(accumulator, new_delta):
        if accumulator is None:
            return new_delta

        if new_delta.content:
            accumulator.content = (
                accumulator.content or ""
            ) + new_delta.content
        if new_delta.role:
            accumulator.role = (accumulator.role or "") + new_delta.role
        if new_delta.tool_calls:
            accumulator.tool_calls = JAImsAgent.__merge_tool_calls(
                accumulator.tool_calls, new_delta.tool_calls
            )

        return accumulator



class JAImsTokensLimitExceeded(Exception):
    """
    Exception raised when the token limit is exceeded.

    Attributes:
        max_tokens -- maximum number of tokens allowed
        messages_tokens -- number of tokens in the messages
        llm_buffer -- buffer for the LLM answer
        has_optimized -- flag indicating if the messages have been optimized
    """

    def __init__(self, max_tokens, messages_tokens, llm_buffer, has_optimized):
        message = f"Max tokens: {max_tokens}\n LLM Answer Buffer: {llm_buffer}\n Messages tokens: {messages_tokens}\n Messages Optimized: {has_optimized}  "
        super().__init__(message)


class JAImsMissingOpenaiAPIKeyException(Exception):
    """
    Exception raised when the OPENAI_API_KEY is missing.
    """

    def __init__(self):
        message = "Missing OPENAI_API_KEY, set environment variable OPENAI_API_KEY or pass it as a parameter to the agent constructor."
        super().__init__(message)


class JAImsOpenAIErrorException(Exception):
    """
    Exception raised when there is an error with OpenAI.

    Attributes:
        message -- explanation of the error
        openai_error -- the error from OpenAI
    """

    def __init__(self, message, openai_error):
        super().__init__(message)
        self.openai_error = openai_error


class JAImsMaxConsecutiveFunctionCallsExceeded(Exception):
    """
    Exception raised when the maximum number of consecutive function calls is exceeded.

    Attributes:
        max_consecutive_calls -- maximum number of consecutive calls allowed
    """

    def __init__(self, max_consecutive_calls):
        message = f"Max consecutive function calls exceeded: {max_consecutive_calls}"
        super().__init__(message)


class JAImsUnexpectedFunctionCall(Exception):
    """
    Exception raised when an unexpected function call occurs.

    Attributes:
        func_name -- name of the unexpected function
    """

    def __init__(self, func_name):
        message = f"Unexpected function call: {func_name}"
        super().__init__(message)



from __future__ import annotations

from enum import Enum
import time
from typing import Any, List, Optional, Dict, Union
import openai
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai import Stream
import tiktoken
import logging
import random
import json

from jaims.function_handler import JAImsFuncWrapper, parse_function_wrappers_to_tools

DEFAULT_MAX_TOKENS = 512
MAX_CONSECUTIVE_CALLS = 10


class JAImsGPTModel(Enum):
    """
    The OPENAI GPT models available.
    Only those that support functions are listed, so just:
    gpt-3.5-turbo-0613, gpt-3-5-turbo-16k-0613, gpt-4-0613
    """

    GPT_3_5_TURBO = ("gpt-3.5-turbo", 4096, 0.0015, 0.002)
    GPT_3_5_TURBO_16K = ("gpt-3.5-turbo-16k", 16384, 0.003, 0.004)
    GPT_3_5_TURBO_0613 = ("gpt-3.5-turbo-0613", 4096, 0.0015, 0.002)
    GPT_3_5_TURBO_16K_0613 = ("gpt-3.5-turbo-16k-0613", 16384, 0.003, 0.004)
    GPT_3_5_TURBO_1106 = ("gpt-3.5-turbo-1106", 16385, 0.001, 0.002)
    GPT_4 = ("gpt-4", 8192, 0.03, 0.06)
    GPT_4_32K = ("gpt-4-32k", 32768, 0.06, 0.12)
    GPT_4_0613 = ("gpt-4-0613", 8192, 0.03, 0.06)
    GPT_4_32K_0613 = ("gpt-4-32k-0613", 32768, 0.06, 0.12)
    GPT_4_1106_PREVIEW = ("gpt-4-1106-preview", 128000, 0.01, 0.03)
    GPT_4_VISION_PREVIEW = ("gpt-4-vision-preview", 128000, 0.01, 0.03)

    def __init__(self, string, max_tokens, price_1k_tokens_in, price_1k_tokens_out):
        self.string = string
        self.max_tokens = max_tokens
        self.price_1k_tokens_in = price_1k_tokens_in
        self.price_1k_tokens_out = price_1k_tokens_out

    def __str__(self):
        return self.string


class JAImsTokensExpense:
    """
    Tracks the number of tokens spent on a job and on which GPTModel.
    """

    def __init__(
        self,
        gpt_model: JAImsGPTModel,
        prompt_tokens=0,
        completion_tokens=0,
        total_tokens=0,
        rough_estimate=False,
    ):
        self.gpt_model = gpt_model
        self.prompt_tokens = prompt_tokens
        self.completion_tokens = completion_tokens
        self.total_tokens = total_tokens
        self.rough_estimate = rough_estimate

    @staticmethod
    def from_openai_usage_dictionary(
        gpt_model: JAImsGPTModel, dictionary: dict
    ) -> JAImsTokensExpense:
        return JAImsTokensExpense(
            gpt_model=gpt_model,
            prompt_tokens=dictionary["prompt_tokens"],
            completion_tokens=dictionary["completion_tokens"],
            total_tokens=dictionary["total_tokens"],
        )

    def spend(self, prompt_tokens, completion_tokens, total_tokens):
        self.prompt_tokens += prompt_tokens
        self.completion_tokens += completion_tokens
        self.total_tokens += total_tokens

    def add_from(self, other_expense: JAImsTokensExpense):
        self.prompt_tokens += other_expense.prompt_tokens
        self.completion_tokens += other_expense.completion_tokens
        self.total_tokens += other_expense.total_tokens
        if other_expense.rough_estimate:
            self.rough_estimate = True  # becomes rough if summed with something rough

    def get_cost(self):
        return (self.prompt_tokens / 1000) * self.gpt_model.price_1k_tokens_in + (
            self.completion_tokens / 1000
        ) * self.gpt_model.price_1k_tokens_out

    def __str__(self):
        string_repr = (
            f"GPT model: {self.gpt_model}\n"
            f"Prompt tokens: {self.prompt_tokens}\n"
            f"Completion tokens: {self.completion_tokens}\n"
            f"Total tokens: {self.total_tokens}\n"
            f"Cost: {round(self.get_cost(),4)}$"
        )

        if self.rough_estimate:
            string_repr += "\n(warning: rough estimate)"

        return string_repr

    def to_json(self):
        return {
            "model": self.gpt_model.string,
            "prompt_tokens": self.prompt_tokens,
            "completion_tokens": self.completion_tokens,
            "total_tokens": self.total_tokens,
            "cost": self.get_cost(),
            "rough_estimate": self.rough_estimate,
        }


def estimate_token_count(string: str, model: JAImsGPTModel) -> int:
    """Returns the number of tokens in a text string."""

    encoding = tiktoken.encoding_for_model(model.string)
    num_tokens = len(encoding.encode(string))
    return num_tokens


class ErrorHandlingMethod(Enum):
    FAIL = "fail"
    RETRY = "retry"
    EXPONENTIAL_BACKOFF = "exponential_backoff"


def __handle_openai_error(error: openai.OpenAIError) -> ErrorHandlingMethod:
    # errors are handled according to the guidelines here: https://platform.openai.com/docs/guides/error-codes/api-errors (dated 03/10/2023)
    # this map indexes all the error that require a retry or an exponential backoff, every other error is a fail
    error_handling_map = {
        openai.RateLimitError: ErrorHandlingMethod.EXPONENTIAL_BACKOFF,
        openai.InternalServerError: ErrorHandlingMethod.RETRY,
        openai.APITimeoutError: ErrorHandlingMethod.RETRY,
    }

    for error_type, error_handling_method in error_handling_map.items():
        if isinstance(error, error_type):
            return error_handling_method

    return ErrorHandlingMethod.FAIL


class JAImsOpenaiKWArgs:
    """
    Represents the keyword arguments for the JAIms OpenAI wrapper.
    This class entirely mirrors the openai API parameters, so refer to it for documentation.
    (https://platform.openai.com/docs/api-reference/chat/create).

    Args:
        model (JAImsGPTModel, optional): The OpenAI model to use. Defaults to JAImsGPTModel.GPT_3_5_TURBO.
        messages (List[dict], optional): The list of messages for the conversation. Defaults to an empty list, it is automatically populated by the run method so it is not necessary to pass them.
        max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 500.
        stream (bool, optional): Whether to use streaming for the API call. Defaults to False.
        temperature (float, optional): The temperature for generating creative text. Defaults to 0.0.
        top_p (Optional[int], optional): The top-p value for nucleus sampling. Defaults to None.
        n (int, optional): The number of responses to generate. Defaults to 1.
        seed (Optional[int], optional): The seed to be passed to openai to have more consistent outputs. Defaults to None.
        frequency_penalty (float, optional): The frequency penalty for avoiding repetitive responses. Defaults to 0.0.
        presence_penalty (float, optional): The presence penalty for encouraging diverse responses. Defaults to 0.0.
        logit_bias (Optional[Dict[str, float]], optional): The logit bias for influencing the model's output. Defaults to None.
        response_format (Optional[Dict], optional): The format for the generated response. Defaults to None.
        stop (Union[Optional[str], Optional[List[str]]], optional): The stop condition for the generated response. Defaults to None.
        tool_choice (Union[str, Dict], optional): The choice of tool to use. Defaults to "auto".
        tools (Optional[List[JAImsFuncWrapper]], optional): The list of function wrappers to use as tools. Defaults to None.
    """

    def __init__(
        self,
        model: JAImsGPTModel = JAImsGPTModel.GPT_3_5_TURBO,
        messages: List[dict] = [],
        max_tokens: int = DEFAULT_MAX_TOKENS,
        stream: bool = False,
        temperature: float = 0.0,
        top_p: Optional[int] = None,
        n: int = 1,
        seed: Optional[int] = None,
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
        logit_bias: Optional[Dict[str, float]] = None,
        response_format: Optional[Dict] = None,
        stop: Union[Optional[str], Optional[List[str]]] = None,
        tool_choice: Union[str, Dict] = "auto",
        tools: Optional[List[JAImsFuncWrapper]] = None,
    ):
        self.model = model
        self.messages = messages
        self.max_tokens = max_tokens
        self.stream = stream
        self.temperature = temperature
        self.top_p = top_p
        self.n = n
        self.seed = seed
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.logit_bias = logit_bias
        self.response_format = response_format
        self.stop = stop
        self.tool_choice = tool_choice
        self.tools = tools

    def to_dict(self):
        kwargs = {
            "model": self.model.string,
            "temperature": self.temperature,
            "n": self.n,
            "stream": self.stream,
            "messages": self.messages,
            "top_p": self.top_p,
            "max_tokens": self.max_tokens,
            "seed": self.seed,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "response_format": self.response_format,
            "stop": self.stop,
        }

        kwargs = {key: value for key, value in kwargs.items() if value is not None}

        if self.logit_bias:
            kwargs["logit_bias"] = self.logit_bias

        if self.tools:
            kwargs["tools"] = parse_function_wrappers_to_tools(self.tools)
            kwargs["tool_choice"] = self.tool_choice

        return kwargs


class JAImsOptions:
    """
    Represents the options for JAImsAgent.

    Args:
        leading_prompts (Optional[List[Dict]]): A list of leading prompts, these will be always prepended to the history for each run.
        trailing_prompts (Optional[List[Dict]]): A list of trailing promptsm, these will be always appended to the history for each run.
        max_consecutive_function_calls (int): The maximum number of consecutive function calls allowed (defaults to 10 to avoid infinite loops).
        optimize_context (bool): Whether to optimize the context in the history manager or not, defaults to True.
        message_history_size (Optional[int]): The size of the message history for each run, only the last n messages will be passed, defaults to none (every message is passed until optimization starts).
        max_retries (int): The maximum number of retries after a failing openai call.
        retry_delay (int): The delay between each retry.
        exponential_base (int): The base for exponential backoff calculation.
        exponential_delay (int): The initial delay for exponential backoff.
        exponential_cap (Optional[int]): The maximum delay for exponential backoff.
        jitter (bool): Whether to add jitter to the delay (to avoid concurrent firing).
        debug_stream_function_call (bool): Prints the arguments streamed by OpenAI during function call when streaming enabled.
    """

    def __init__(
        self,
        leading_prompts: Optional[List[Dict]] = None,
        trailing_prompts: Optional[List[Dict]] = None,
        max_consecutive_function_calls: int = MAX_CONSECUTIVE_CALLS,
        optimize_context: bool = True,
        message_history_size: Optional[int] = None,
        max_retries=15,
        retry_delay=10,
        exponential_base: int = 2,
        exponential_delay: int = 1,
        exponential_cap: Optional[int] = None,
        jitter: bool = True,
        debug_stream_function_call=False,
    ):
        self.leading_prompts = leading_prompts
        self.trailing_prompts = trailing_prompts
        self.max_consecutive_function_calls = max_consecutive_function_calls
        self.optimize_context = optimize_context
        self.message_history_size = message_history_size
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self.exponential_base = exponential_base
        self.exponential_delay = exponential_delay
        self.exponential_cap = exponential_cap
        self.jitter = jitter
        self.debug_stream_function_call = debug_stream_function_call


def get_openai_response(
    openai_kw_args: JAImsOpenaiKWArgs,
    call_options: JAImsOptions,
) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
    retries = 0
    logger = logging.getLogger(__name__)
    # keeps how long to sleep between retries
    sleep_time = call_options.retry_delay
    # keeps track of the exponential backoff
    backoff_time = call_options.exponential_delay

    # print(json.dumps(openai_kw_args.messages, indent=4))

    while retries < call_options.max_retries:
        try:
            response = openai.chat.completions.create(
                **openai_kw_args.to_dict(),
            )

            return response
        except openai.OpenAIError as error:
            logger.error(f"OpenAI API error:\n{error}\n")
            error_handling_method = __handle_openai_error(error)

            if error_handling_method == ErrorHandlingMethod.FAIL:
                raise Exception(f"OpenAI API error: {error}")

            if error_handling_method == ErrorHandlingMethod.RETRY:
                sleep_time = call_options.retry_delay

            elif error_handling_method == ErrorHandlingMethod.EXPONENTIAL_BACKOFF:
                logger.info(f"Performing exponential backoff")
                jitter = 1 + call_options.jitter * random.random()
                backoff_time = backoff_time * call_options.exponential_base * jitter

                if (
                    call_options.exponential_cap is not None
                    and backoff_time > call_options.exponential_cap
                ):
                    backoff_time = call_options.exponential_cap * jitter

                sleep_time = backoff_time

            logger.warning(f"Retrying in {sleep_time} seconds")
            time.sleep(sleep_time)
            retries += 1

    max_retries_error = f"Max retries exceeded! OpenAI API call failed {call_options.max_retries} times."
    logger.error(max_retries_error)
    raise Exception(max_retries_error)



from setuptools import setup, find_packages

# Read requirements.txt
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="jaims-py",
    version="1.0.1",
    packages=find_packages(),
    description="A Python package for creating simple AI Agents using the OpenAI API.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Marco Musella",
    url="https://github.com/dev-mush/jaims-py",
    license="MIT",
    keywords="OpenAI GPT-3 and GPT-4 function enabled agent",
    install_requires=requirements,
)
